/****************************************************************************/
/*                 MASS project: TURBINA module file                        */
/*                            scan.c                                        */
/*                    Description is given in scan.h                        */
/* Editor tabular size: 4                                                   */
/****************************************************************************/

/* Insert the definition SCANTEST in compilation for the module self-tests:
gcc -c scan.c .... -DSCANTEST=1 */
//#define SCANTEST

#include "scan.h"

/*---------------------------------------------------------------------*/
/*--- Public static data storages for accumulation of scans -----------*/
/*----- they are allocated by sc_init() and released be sc_done() -----*/
/** Accumulated forward scans [Nchan][Ncnt]*/
static double ** sn_fscan = NULL ;
/** Accumulated backward scans */
static double ** sn_bscan = NULL ;
/** Merged accumulated scans */
static double ** sn_scan = NULL ;

/** Number of channels for which the scan storages are allocated */
static int sn_nchan = 0 ;
/** Number of elements in each elementary scan */
static int sn_ncnt = 0 ;
/** Current numbers of accumulated scans in channels */
static int * sn_curfscan = NULL ;
static int * sn_curbscan = NULL ;

#ifdef SCANTEST
/*---------------------------- TEST PART ---------------------------------*/
#define NTEST 100

/*----- Comes normally from configuration file: cf_... vars and defs --------*/

/* Apertures: */
/* number of apertures in segmentator : */
#define CF_NCHAN  4
#define CF_NSCAN   25
/* Physical size of segmentator annular apertures [cm] : */
double cf_aper_inn[] = { 0.000, 0.102, 0.188, 0.354};
double cf_aper_out[] = { 0.100, 0.185, 0.350, 0.650};
/* Exposure section: */
#define CF_MICROEXP  (0.004) 	/* < microexposure: times per scan count */
#define CF_NCNT 100 	/* number of microexposures (or microsteps) per scan */
/* Magnification of the system: */
double cf_magnif = 20.0 ;
#define SCALE	1.5		/* Focal plane scale, in arcsec/microstep */
const double cf_bkgr[] = {30., 32., 40., 70.}; /* background counts measured in
channels per second */

#define CF_DEFRANGE 12	/* pixels from the center (see sn_focus() docs)*/
/*------------- Other supplementary info: ------------------------*/
double sys_qe = 0.05 ;	/* System quantum efficiency */
double flux0 = 1E6 ; 	/* Photons per second per cm^2 from the mag=0 A0-star*/

/* Coordinates of simulated crossing the edges of the triangle aperture */
const double x10 = 0.25*CF_NCNT, x20 = 0.65*CF_NCNT ;
#define TAUWANDER	0.5	/* Characteristic time [sec] of image wandering */
/* Number of scanning microsteps per wandering characteristic time: */
const int steperwan = (int)(TAUWANDER / CF_MICROEXP + 0.5) ;

#define PI_NUMB (3.1415926535897932) 
#define ERF05	(0.476936)	/* Error function argument when it is equal to 0.5*/
/** 
	@brief	Image wandering simulation in one dimention
	@param	correlength	duration of correlation of deviations in calls
	@param	sigma_wander	RMS deviation of wandering
	@param	idum	random seed
	@return	random wandering deviation
	
	Function simulates the process of stellar image wandering in one dimention
	using the RMS deviation and the time span in units of function calls at
	which the deviations are correlated. This span equals to the characteristic
	time of wandering divided by the time interval with which the function is
	called.
	
	Initialize the function by the call with idum=0, use it, and then
	deinitialize by the call again with idum=0.
*/
double wanderdev(const int correlength, 
				const double sigma_wander,
				int *idum) {
static int n  = 0, len = 0 ;
static double fact = 0;
double dev ;
int i ;
static double * buf = NULL ;

	if ((*idum==0) || (len==0)) {
		if (buf==NULL) {
			/* Initialize: */
			len = correlength ;
			/** dispersion of \em len coadded normal-distributed 
				randoms with unity dispersion */
			fact = sqrt((double)len) ; 
			/* Allocate the buffer with for len randoms */
			buf = dvector(0,len-1);
			/* fill in this buffer: */
			for (n=0; n<len; n++) {
				buf[n] = gasdev(&(*idum)) * sigma_wander / fact ;
//				printf ("%f\n",buf[n]) ;
			}
		} else {
			/* deallocate: */
			free_dvector(buf,0,len-1) ;
			buf=NULL; n=len=0;
			return 0 ;
		}
	} else {
		n++;
		n = n % len ;
		buf[n] = gasdev(&(*idum)) * sigma_wander / fact ;
	}
	/* Compute the deviation: */
	dev = 0 ;
	for (i=0; i<n; i++)
		dev += buf[i] ;
	return dev ;
}

/* Define this to view instant scans with xmgrace: */
//#define VIEWSCAN
/* Define this to view merged scans with xmgrace: */
//#define VIEWMSCAN

int main( int ac, char **av ) {

/* Receiving buffers for channels counts: */
count_t channelA[CF_NCNT] ;
count_t channelB[CF_NCNT] ;
count_t channelC[CF_NCNT] ;
count_t channelD[CF_NCNT] ;
count_t * channel[] = { (count_t*)(&channelA), (count_t*)(&channelB), 
					(count_t*)(&channelC), (count_t*)(&channelD) } ;

#define SCALE 1.5	/* 1.5 arcsec per microstep */
#define ANGLE 0.707107	/* 45 degree angle between the edge and ardius-vector */

double area, flux, 
	magnitude = 0, 
	sigma_wander = 1, /* arcsec */
	seeing = 1, /* arcsec */ 
	sigma_scint = 0.1, 
	defocus = 0, 
	x1, 
	factor, mean, dev, dx, x,
	shx, shy,
	cx1, cx2, cdefocus;

int i, j, idum=(-13), err, test, meas, which;

if (ac<2) {
	printf(" MASS project: SCAN module test\n");
	printf(" Use: %s <star magnitude> <wander> <seeing> <scint> <defocus>\n",av[0]);
	printf(" where image wandering and seeing are in arcseconds \n");
	printf("   and scintillation is a relative flux variation (<1) \n");
	return -1 ;
}

magnitude = atof(av[1]) ;
if (ac>2) { x1 = atof(av[2]) ; if (x1>0) sigma_wander=x1; }
if (ac>3) { x1 = atof(av[3]) ; if (x1>0) seeing=x1; }
if (ac>4) { x1 = atof(av[4]) ; if (x1>0) sigma_scint=x1; }
if (ac>5) { x1 = atof(av[5]) ; if (x1>0) defocus=x1; }
if ((magnitude<-2)) {
	printf(" Error: magnitude is >=-2 \n");
	return -1 ;
}

/* Cycle of all the system work scenario repetitions : */

for (test=0; test<NTEST; test++) {

//	printf("#Initialize the module.\n") ;
	err = sn_init(CF_NCNT,CF_NCHAN); 
	if (err) { fprintf(stderr,sc_error(err)); return err;}
	
	for (meas=0; meas<CF_NSCAN*2; meas++) {
		/* Direction sign : */
		which = 1-2*(meas%2) ;
		for (i=0; i<CF_NCHAN; i++) {
			mean = 0 ;
			cx1 = cf_magnif*cf_aper_out[i]/2 ;
			cx2 = cf_magnif*cf_aper_inn[i]/2 ;
			area = (cx1*cx1-cx2*cx2) * PI_NUMB ;
			flux = flux0 * area * pow(10.0,-0.4*magnitude) * sys_qe ;
			for (j=0; j<CF_NCNT; j++) {
				dev = wanderdev(steperwan,sigma_wander,&idum);
				/* which edge is closer? */
				if (which>0)
					x = (double)j;
				else
					x = (double)(CF_NCNT-j-1) ;
				cx1 = x - (x10+dev) ;
				cx2 = (x20+dev) - x ;
				/* Coordinate difference with the edge: */
				dx = (fabs(cx1)<fabs(cx2)) ? cx1 : cx2 ;
				/* Stellar profile fraction which is seen: */
				factor = 0.5 + erf(dx/(seeing/SCALE) * ERF05) / 2.0 ;
				/* Scintillation: */
				dev = pow(10.0,0.4*gasdev(&idum)*sigma_scint) ;
//				printf("%d %f %f %f %f\n",(int)x,dev,dx,factor,dev);
				factor *= dev;
//				printf("%d %f\n",(int)x,factor);
				channel[i][j] = (count_t)(0.5+
					poidev((flux*factor + cf_bkgr[i])*CF_MICROEXP, &idum)) ;
				mean += (double)channel[i][j] ;
			}

			mean /= CF_NCNT ;
			if ((err = sn_accum(channel[i], i, which))<0) {
				fprintf(stderr,sc_error(err)); return err; }
		}
		#ifdef VIEWSCAN
		for (i=0; i<CF_NCHAN; i++) 
			for (j=0; j<CF_NCNT; j++) {
				if (which>0)
					x = (double)j;
				else
					x = (double)(CF_NCNT-j-1) ;
				printf("%d %d\n",(int)x,channel[i][j]);
			}
		#endif
	}

	/* Average scans: */
#ifndef VIEWSCAN
	printf("\n# Average scans");
	if (defocus!=0) printf(" and compute the focus:\n");
	else printf(" and compute the edge positions:\n");
#endif	
	if ((err = sn_merge())<0) { fprintf(stderr,sc_error(err)); return err; }
#ifdef VIEWMSCAN
	printf("# Ok. Here they are:\n# x");

	for (i=0; i<CF_NCHAN; i++) printf("\t%s",sc_apername(CF_NCHAN,i));	
	printf("\n");
	for (j=0; j<CF_NCNT; j++) {
		printf("%d",j);
		for (i=0; i<CF_NCHAN; i++) 
			printf("\t%d",(int)((sn_get(i,0))[j]+0.5)) ;
		printf("\n");
	}
#endif
	for (i=0; i<CF_NCHAN; i++) {
		if (defocus==0) {
			if ((err = sn_center (sn_get(i,0), CF_NCNT, &cx1, &cx2))<0) {
				fprintf(stderr,sc_error(err)); return err; }
			printf("# Chan %s: x1=%f, x2=%f i.e. ",sc_apername(CF_NCHAN,i),
				cx1, cx2);
			if ((err = sn_shift (sn_get(i,0), CF_NCNT, SCALE, ANGLE,
				x10, x20, &shx, &shy))<0) {
				fprintf(stderr,sc_error(err)); return err; }
			printf("sh_x=%5.2f\", sh_y=%5.2f\"\n", shx, shy);
		} else {
			if ((err = sn_focus (sn_get(i,0), CF_NCNT, CF_DEFRANGE, &cdefocus))<0) {
				fprintf(stderr,sc_error(err)); return err; }
			printf("# Chan %s: defocus=%f\n",sc_apername(CF_NCHAN,i), cdefocus);
		}
	}
	printf("# Release module arrays:\n") ;
	sn_done() ;
	
	printf("#===================== %d ===================\n",test);

} /* of test-cycle */

idum = 0;
dev = wanderdev(0,0,&idum);
printf("# Ok.\n");

return 0 ;
} /* end of main() */
#endif


/*---------------------------- sc_init ----------------------------------*/

int sn_init(int ncnt, int nchan) {

	int i, j ;
	
	if ((sn_bscan!=NULL) || (sn_fscan!=NULL) || (sn_scan!=NULL) ||
		(sn_curfscan!=NULL) || (sn_curbscan!=NULL))	return SC_ERNNL ;

	if ((ncnt<=0) || (nchan<=0)) return SC_ERPAR ;

	if ((sn_bscan = dmatrix(0, nchan-1, 0, ncnt-1)) == NULL) return SC_ERMEM ;
	if ((sn_fscan = dmatrix(0, nchan-1, 0, ncnt-1)) == NULL) return SC_ERMEM ;
	if ((sn_scan = dmatrix(0, nchan-1, 0, ncnt-1)) == NULL) return SC_ERMEM ;
	
	if ((sn_curfscan = ivector(0,nchan-1))==NULL) return SC_ERMEM ; 
	if ((sn_curbscan = ivector(0,nchan-1))==NULL) return SC_ERMEM ; 

	/* Initialize scans with zeroes: */
	for (j=0; j<nchan; j++)  {
		sn_curfscan[j] = sn_curbscan[j] = 0 ;
		for (i=0; i<ncnt; i++) 
			sn_bscan[j][i] = sn_fscan[j][i] = sn_scan[j][i] = 0 ;
	}

	sn_ncnt = ncnt ;
	sn_nchan = nchan ;
	
	return 0 ;
}

/*---------------------------- sn_done ----------------------------------*/

void sn_done() {
	if (sn_bscan!=NULL) 
		{ free_dmatrix(sn_bscan,0,sn_nchan-1,0,sn_ncnt-1); sn_bscan = NULL ;}
	if (sn_fscan!=NULL) 
		{ free_dmatrix(sn_fscan,0,sn_nchan-1,0,sn_ncnt-1); sn_fscan = NULL ;}
	if (sn_scan!=NULL) 
		{ free_dmatrix(sn_scan,0,sn_nchan-1,0,sn_ncnt-1); sn_scan = NULL ;}
	if (sn_curfscan!=NULL) 
		{ free_ivector(sn_curfscan,0,sn_nchan-1); sn_curfscan = NULL ;}
	if (sn_curbscan!=NULL) 
		{ free_ivector(sn_curbscan,0,sn_nchan-1); sn_curbscan = NULL ;}
}

/*---------------------------- sn_accum ---------------------------------*/

int sn_accum(count_t * scan, int chan, int which) {
	int i ;
	if (scan==NULL) return SC_ERNUL ;
	if ((chan<0) || (chan>=sn_nchan)) return SC_ERPAR ;
	if (which==SN_FORW) {
		/* forward scan: */
		for (i=0; i<sn_ncnt; i++)
			sn_fscan[chan][i] += (double)(scan[i]) ;
		(sn_curfscan[chan])++ ;
	} else {
		if (which==SN_BACK) {
			/* backward scan: */
			for (i=0; i<sn_ncnt; i++)
				sn_bscan[chan][i] += (double)(scan[i]) ;
			(sn_curbscan[chan])++ ;
		} else 
			return SC_ERPAR ;
	}
	return 0 ;
}

/*---------------------------- sn_merge ---------------------------------*/

int sn_merge() {
	int i, chan, tot ;
	
	if (sn_scan==NULL) return SC_ERNUL ;
	for (chan=0; chan<sn_nchan; chan++) {
		tot = sn_curfscan[chan] + sn_curbscan[chan] ;
		if (tot>0) {
			for (i=0; i<sn_ncnt; i++)
				sn_scan[chan][i] += sn_fscan[chan][i] / tot ;
			for (i=0; i<sn_ncnt; i++)
				sn_scan[chan][sn_ncnt-1-i] += sn_bscan[chan][i] / tot ;
		}
	}
	return 0 ;
}

/*---------------------------- sn_get -----------------------------------*/

double * sn_get(int chan, int which) {
	double * scan ;
	scan = NULL ;
	if ((chan>=0) && (chan<sn_nchan))
		switch (which) {
			case SN_MERG:
				scan = sn_scan[chan] ; break ;
			case SN_FORW:
				scan = sn_fscan[chan] ; break ;
			case SN_BACK:
				scan = sn_bscan[chan] ; break ;
		}
	return scan ;
} 

/*---------------------------- sn_center --------------------------------*/

int sn_center(double * scan, int ncnt, double * x1, double * x2) {
#define K_BKGR 5	/* least value corresponds to K_BKGR=1 */
#define K_PLAT0	10	/* crude plateau estimate: K_PLAT0-th largest */
#define UP_LIN 0.75	/* upper boundary of the linear range of slope */
#define LO_LIN 0.25	/* lower boundary of the linear range of slope */
	int i, j ;
	double bkgr, plat, level ;
	double * sorscan ;

	if (scan==NULL) return SC_ERNUL	;
	/* Sort the copied scan: */
	if ((sorscan = dvector(0,ncnt-1))==NULL) return SC_ERMEM ;
	for (i=0; i<ncnt; i++) sorscan[i] = scan[i] ;
	sort(ncnt,sorscan-1) ; 
	/* Find the background level as K_BKGR-th smallest count: */
	bkgr = sorscan[K_BKGR-1] ;
	/* and plateau level: */
	plat = sorscan[ncnt-1-K_PLAT0] ;
	
	/* True plateau level is a median of counts which are above UP_LIN fraction
	of the intensity range: */
	level = bkgr + UP_LIN*(plat-bkgr) ;
	plat = 0 ;
	for (i=K_BKGR-1; i<ncnt-K_PLAT0; i++) 
		if (sorscan[i]>level) {
			j = i+(ncnt-K_PLAT0-i)/2 ;
			plat = sorscan[j] ;
			break ;
		}
	free_dvector(sorscan,0,ncnt-1);
	if (plat==0)
		/* No plateau found */
		return SC_ERNOD ;
	/* 50% level: */
	level = (plat + bkgr) / 2 ;
	/* search for the level crossing while the star appears: */
	*x1 = 0 ;
	for (i=0; i<ncnt-1; i++) {
		if ((scan[i]<=level) && (scan[i+1]>level)) {
			*x1 = (double)i + (level-scan[i])/(scan[i+1]-scan[i]) ;
			break ;
		}
	}
	/* search for the level crossing where the star disappears: */
	*x2 = 0 ;
	for (i=ncnt-2; i>=0; i--) {
		if ((scan[i]>=level) && (scan[i+1]<level)) {
			*x2 = (double)i + (scan[i]-level)/(scan[i]-scan[i+1]) ;
			break ;
		}
	}

//	printf ("sc[%i]=%f j=%d, plat=%f\n",i,sorscan[i],j,plat);
	return 0 ;
}

/*---------------------------- sn_shift ---------------------------------*/

int sn_shift(double * scan, int ncnt, double scale, double angle, 
double x1_0, double x2_0, double * shx, double * shy) {
#define OUTFOV (+1000)
	int err, isout ;
	double x1 = 0, x2 = 0, xc, xc0 ;

	
	if ((err=sn_center(scan,ncnt,&x1,&x2))<0) return err ;
	
	/* Radial shift calculations: */
	isout = ((x1==0) || (x2==0)) ;
	if (!isout) {
		/* Y-Shift is computed from the change of the length of the lighted-area
		of the scan */
		xc = (x2-x1) ;
		xc0 = (x2_0 - x1_0) ;
		*shy = scale * ((xc - xc0)/2) * tan(angle) ;
		/* X-shift is made from the scaled displacement of center of
		lighted-area of the scan */
		xc = (x1+x2)/2 ;
		xc0 = (x1_0+x2_0)/2 ;
		*shx = (xc-xc0) * scale ;
		return 0 ;
	} else {
		*shy = *shx = OUTFOV ;
		if (x1!=0) 
			*shx = (x1-x1_0) * scale ;
		if (x2!=0) 
			*shx = (x2-x2_0) * scale ;
		return 0 ;
	}
#undef OUTFOV
}

/*---------------------------- sn_focus ---------------------------------*/

int sn_focus(double * scan, int ncnt, int defrange, double * defocus) {

	double peak, d1, d2 ;
	int i, xpeak, imin, imax, n ;
	if (scan==NULL) return SC_ERNUL	;
	if (defrange<1) return SC_ERPAR ;
	/* peak location: */
	peak = 0;
	for (i=0; i<ncnt-1; i++) {
		if (scan[i]>peak) {
			peak = scan[i];
			xpeak = i;
		}
	}
	if (xpeak>=ncnt-1) return SC_ERNOD ;
	
	imin = ((xpeak>defrange) ? (xpeak - defrange) : 0) ;
	imax = (((xpeak+defrange) >= ncnt) ? (ncnt-1) : (xpeak + defrange)) ;
	d1 = d2 = 0 ;
	n = (imax-imin+1) ;
	for (i=imin ; i<=imax; i++) {
		d1 += scan[i] / n ;
		d2 += scan[i] * scan[i] / n ;
	}
	if (d1<=0) return SC_ERNOD ;
	*defocus = d2 / (d1*d1) ;
	return 0 ;
}	
