/*--------------------------------------------------------------------------*/
/** @file scan.h
*                  MASS project: TURBINA module file                       
*                       header file for scan.c                             
*                                                                          
* The module SCAN contains a set of utilities which handle the scans obtained
* in any servo-regime of MASS device which involves synchronous counting 
* simultaneous with moving of aperture disk (i.e. \em scanning). Here 
* "apertures" are the focal plane diaphragms or lens as opposed to entrance 
* apertures of the system which are played around in SCIND and WEIF module. 
* Servo-regimes of MASS which are handled here are focusing and centering of 
* the MASS device apertures.
*
* The names of public members in this module start with the "sn_" prefix.
*
* \b Usage:
*
* Double-type arrays which are aimed to accumulate the scanning results are
* private global arrays in this module. This is made so to avoid the memory
* handling and related potential errors while using the module in a big and
* complicated program. Thus, before starting the scanning process, the sn_init()
* is to be called. Each time the new scan is received, the scan accumulation
* utility sn_accum() is called with the proper value of the flag of scanning 
* direction. After finishing the scanning, the utility sn_merge() should be
* called to obtain the resulting average scan where forward and backward scans 
* are properly coadded and normalized. The utility sn_get() returns the pointer 
* to one of the scans (forward, backward or merged). The "scientific" functions
* sn_center() and sn_focus() extract the information on the position of the star
* in the centering triangle aperture and on the focus displacement, respectively. And, finally, the cleaning of
*
* The scans are made of individual counts from the MASS device channels and
* represent the \em count_t-type arrays. The latter type definition is 
* borrowed from the module SCIND.
*
* @author N. Shatsky, Sternberg Institute (kolja@sai.msu.ru)               
*
* @version 1.0: Creation                                                    */
/* This header file is equipped with documentation made in comments in a
* format suitable for production of LaTeX or HTML documentation with the
* doxygen software.
* Editor tabular size: 4                                                    */
/****************************************************************************/

#ifndef _SCAN_H_
#define _SCAN_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "scind.h"

#define SN_FORW	(+1) /**< \em which -flag to access forward scan */
#define SN_BACK	(-1) /**< \em which -flag to access back-scan (see sn_get()) */
#define SN_MERG	0    /**< \em which -flag to access merged scan */
     
/****************************************************************************/
/*                     Declarations of local functions:                     */
/****************************************************************************/

/**
	@brief	Initialize the scan arrays before accumulation of scans started
	@param	ncnt	Number of counts in each scan (forward or backward)
	@param	nchan	Number of channels (entrance apertures) where the scanning 
					will go on
	@return	0 if Ok or <0 on error
	
	The module contains the arrays \em sn_fscan, \em sn_bscan and \em
	sn_scan which are aimed to, respectively, accumulate forward and back scans
	in all channels and to produce the merged scan. These are three matrices 
	[Nchannel][Ncnt] where each row is a space to accumulate the scan in one
	channel. They all are allocated here.
	
	The dimentions of allocated arrays are determined by the numbers \em ncnt
	and \em nchan which are copied in private module global
	variables with the same names and prefix \em sn_. Other private global
	variables - \em sn_curfscan and \em sn_curbscan - count the accumulated
	scans in \em sn_fscan and \em sn_bscan, respectively.

	This function should be called, obviously, before the first scanning pass is
	finished.

	@see sn_accum() sn_done()
*/
int sn_init(int ncnt, int nchan) ;

/**
	@brief	Deallocate the memory reserved for module scan arrays
	@return	void
	
	This function releases the memory which is allocated for forward, back and
	merged scans by sn_init() according to the module variable \em sn_nchan.

	\note
	This function must be kept in strict sync with sn_init().
*/
void sn_done() ;

/**
	@brief	Add the current scan to the respective scan storage
	@param	scan	pointer to the \em count_t -type scan array
	@param	chan	channel number (0,1..)
	@param	which 	Direction flag: SN_FORW: forward, SC_BACK: backward 
	@return 0 if Ok, <0 if error
	
	Function adds the given array to the forward and backward scans, latter with inversed
	sequence of indices, and divides resulting values by the sum of number of
	accumulated scans in both directions.
	
	\note
	The length of the scan MUST be the same as those used in sn_init() call.
	
	@see sn_init()
*/
int sn_accum(count_t * scan, int chan, int which) ;

/**
	@brief	Merge back- and forward scans and normalize the result
	@return 0 if Ok, <0 if error
	
	Function coadds the forward and backward scans in all channels, using the
	backward scans inversed; and divides resulting values by the sum of
	number of accumulated scans in both directions.

	@see sn_accum()
*/
int sn_merge() ;

/**
	@brief Return the pointer to the needed scan
	@param	chan	channel number (0,1..)
	@param	which SN_BACK for backward scan , SN_FORW for forward scan,
		SN_MERG for merged scan
	@return pointer to the scan array (freed ONLY by sn_done()) 

*/
double * sn_get(int chan, int which) ;

/**
	@brief	Locate the coordinates where the star crosses the edges of a
	centering triangle aperture
	@param	scan	pointer to the \b double-type accumulated scan (use
	sn_get())
	@param	ncnt	length of the scan
	@param	x1	Resulting coordinate of the crossing point of the \em left 
	triangle edge [microstep, i.e. pixel of scan]
	@param	x2	Resulting coordinate of the crossing point of the \em right 
	triangle edge [microstep]
	@return 0 if Ok, <0 if error
	
	Function searches for the positions on the scan where the scan counts cross 
	the 50\%-level of the light variation range: the points where the star 
	appeares from and disappeares behind the edges of the triangle aperture.
	
	The method is based on a linear interpolation of points on the very slopes
	of the curve. The points are taken which surround the middle of slopes.
	
	Under average conditions (seeing and wandering of images = 1'',
	scintillations of 10\%, see testing main() of the module) the
	precision of centering is about 0.1 microsteps.

	\note
	
	If one of the crossing points is not covered by the aperture edge while
	scanning, the respective coordinate is returned as zero.
*/
int sn_center(double * scan, int ncnt, double * x1, double * x2) ;

/**
	@brief	Compute the shifts of the star relative to the center of the field 
	of view
	@param	scan	pointer to the \b double-type accumulated scan (use
	sn_get())
	@param	ncnt	length of the scan
	@param	scale	Focal plane scale in [arcsec/microstep] (microstep is a
	"pixel" of scan)
	@param	angle	[radian] between the working edge of the triangle aperture 
	and	radius-vector from the wheel center to its central point
	@param	x1_0	position of appearance of a star on the scan when the star
	is not shifted (as returned by sn_center())
	@param	x2_0	position of disappearance of a star on the scan when the 
	star is not shifted (as returned by sn_center())
	@param	shx 	Resulting shift in transversal direction [arcsec]
	@param	shy 	Resulting shift in radial direction [arcsec] 
	@return 0 if Ok, <0 if error (it is the sn_center() result)
	
	This function is used as an interface to sn_center() to obtain the final
	errors of centering in arcseconds. It thus interprets the scan positions of
	appearance and disappearance of a star in the triangle aperture in order to
	compute the shifts of a star relative to the center of the field of view.
	The triangle aperture is assumed to point by its angle to the axis of the
	apertures wheel. The calculations are made assumed the size of aperture is
	much less than its distance to the center of this wheel.
	
	\notes	
	
	- In case when the scanning limits do not cover one of the points of
	appearance or disappearance of the star, the function returns only the shift
	in X-direction (transversal, i.e. along the scanning direction) which is
	computed in proposition of Y-shift equal to zero; Y-shift \em shy is 
	returned as +1000. If both crossing points are not derivable, both \em shx
	and \em shy are returned as +1000.
	- Desired signs of shifts may be obtained by inversion of signs of the 
	parameters \em scale and/or \em angle.

	@see	sn_center()	
*/
int sn_shift(double * scan, int ncnt, double scale, double angle, 
double x1_0, double x2_0, double * shx, double * shy) ;

/**
	@brief	Compute the focus position dependent measure from the scan 
	obtained in generalized mode
	@param	scan	pointer to the \b double-type accumulated scan (use
	sn_get())
	@param	ncnt	length of the scan [microsteps]
	@param	defrange	Radius of defocus derivations around the scan
	peak [microsteps]
	@param	focus	Resulting measure of defocusing
	@return 0 if Ok, <0 if error
	
	Function computes some integral characteristics of the single accumulated 
	scan which depends significantly on the defocussing of the system.
	
	In current implementation, this characteristic, which is returned in the
	\em defocus variable, is the ratio of the mean square of the signal to
	the squared mean of the signal. The signal is taken from the scan at 
	the positions closer than \em defrange to the peak of the scan.
*/
int sn_focus(double * scan, int ncnt, int defrange, double * focus) ;

#endif
